#include "AutoFit2.h"

#include "MathStats.h"
#include "MathNormal.h"
#include "Kinship.h"

#ifdef __CHROMOSOME_X__
#include "KinshipX.h"
#endif

#include <math.h>

bool AutoFit::useCovariates = false;
bool AutoFit::useProbands = false;

AutoFit2::AutoFit2() { }

void AutoFit2::FitPolygenicModels(Pedigree & ped, bool quiet)
   {
   // Reset previous estimates
   means.Dimension(ped.traitCount); means.Set(_NAN_);
   variances.Dimension(ped.traitCount); variances.Set(_NAN_);
   heritabilities.Dimension(ped.traitCount); heritabilities.Set(_NAN_);

   // Array of informative individuals for each family
   IntArray pheno;

   Kinship kin;

#ifdef __CHROMOSOME_X__
   KinshipX kinX;
#endif

   int probandStatus = useProbands ? ped.affectionNames.SlowFind("proband") : -1;

   if (!quiet)
      {
      // Print header line
      printf("FITTED MODELS%s\n",
             useCovariates ? " (for covariate adjusted residuals)" : "");
      printf("======================================================\n"
             "%15s %12s %12s %12s\n",
             "Trait", "Mean", "Variance", "Heritability");
      }

   // Loop through traits in the pedigree
   for (int t = 0; t < ped.traitCount; t++)
      {
      // First list phenotyped individuals 
      // If we are using covariates, we only consider individuals
      // for which all covariates have been recorded
      pheno.Dimension(0);

      for (int i = 0; i <= ped.count; i++)
         if (ped[i].isPhenotyped(t) && (!useCovariates || ped[i].isFullyControlled()))
            pheno.Push(i);

      if (pheno.Length() == 0)
         {
         if (!quiet)
            printf("Trait: %s -- No informative individuals\n\n",
                   (const char *) ped.traitNames[t]);
         continue;
         }

#ifdef __CHROMOSOME_X__
      int vc_count = 3;
#else
      int vc_count = 2;
#endif

      // The normal set class is our workhorse
      NormalSet mvn;
      mvn.Dimension(1, vc_count);

      // Index to the normal equations within NormalSet
      int index = 0;

      // Build model with polygenes and environment only
      int count = pheno.Length();

      // Phenotypes
      mvn[index].scores.Dimension(count);
      for (int i = 0; i < count; i++)
          mvn[index].scores[i] = ped[pheno[i]].traits[t];

      // Total number of fixed effects
      int fixedEffects = 1 + (useCovariates ? ped.covariateCount : 0);

      // Setup matrix of fixed effects for each family
      mvn[index].linearModel.Dimension(count, fixedEffects);
      for (int i = 0; i < count; i++)
         {
         // Constant for regressing grand mean
         mvn[index].linearModel[i][0] = 1.0;

         // User specified covariates
         if (useCovariates)
            for (int j = 1; j <= ped.covariateCount; j++)
                 mvn[index].linearModel[i][j] = ped[pheno[i]].covariates[j-1];
         }

      // Setup non-shared variances
      mvn[index].varComponents[0].Dimension(count, count);
      mvn[index].varComponents[0].Identity();

      // Setup polygenic variances
      mvn[index].varComponents[1].Dimension(count, count);
      ## MUST INSERT LOOKUP FOR CORRECT KINSHIP ##
      kin.Setup(*ped.families[f]);

      for (int i = 0; i < count; i++)
          for (int j = i; j < count; j++)
              mvn[index].varComponents[1][i][j] =
              mvn[index].varComponents[1][j][i] =
                 ## MUST INSERT LOOKUP FOR CORRECT AUTOSOME KINSHIP ##
                 2.0 * kin(ped[pheno[i]], ped[pheno[j]]);

#ifdef __CHROMOSOME_X__
      // Setup polygenic variances for X chromosome
      mvn[index].varComponents[2].Dimension(count, count);
      ## MUST INSERT LOOKUP FOR CORRECT KINSHIP ##
      kinX.Setup(*ped.families[f]);

      for (int i = 0; i < count; i++)
         for (int j = i; j < count; j++)
             mvn[index].varComponents[2][i][j] =
             mvn[index].varComponents[2][j][i] =
                ## MUST INSERT LOOKUP FOR CORRECT X KINSHIP ##
                2.0 * kinX(ped[pheno[i]], ped[pheno[j]]);
#endif

      // Fit polygenic model
      mvn.Solve();

      double mean = useCovariates ? 0.0 : mvn.means[0];
      double variance = mvn.variances.Sum();
      double heritability = mvn.variances[vc_count - 1] / variance;

      if (!quiet)
         printf("%15s %12.5f %12.5f %12.5f\n",
                (const char *) ped.traitNames[t], mean, variance, heritability);

      means[t] = mean;
      variances[t] = variance;
      heritabilities[t] = heritability;

      if (!useCovariates) continue;

      // Evaluate model to update residuals for each family
      mvn.Evaluate();

      index = 0;

      // First wipe out all the phenotypes, to ensure that we
      // don't get residuals and covariates mixed up ...
      for (int i = 0; i < ped.count; i++)
         ped[i].traits[t] = _NAN_;

      // Replace phenotypes with residuals
      int count = pheno.Length();

      // Replace phenotypes with residuals
      for (int i = 0; i < count; i++)
         ped[pheno[i]].traits[t] = mvn[index].residuals[i];
      }
   }


